/*********************************************************
## This file is part of Prodog.  All files in the distribution of Prodog
## are Copyright 2000 by
##
##     Alva L.  Couch,
##     Associate Professor of Computer Science,
##     Tufts University, Medford MA 02155
##     email: couch@eecs.tufts.edu
##     phone: 617-627-3674
##
## All rights reserved.
##
## Redistribution, modification, and use are permitted provided that this 
## entire copyright notice is duplicated in all such copies, and that any
## documentation, announcements, and other materials related to such
## distribution and use acknowledge that the software was developed by
## Alva L.  Couch at Tufts University in Medford, Massachusetts, USA.
## No charge, other than an "at-cost" distribution fee, may be charged
## for copies, derivations, or distributions of this material without
## the express written consent of the copyright holder.  Neither the
## name of the University nor the names of the authors may be used to
## endorse or promote products derived from this material without
## specific prior written permission.  THIS SOFTWARE IS PROVIDED ``AS
## IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
## WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
## FITNESS FOR ANY PARTICULAR PURPOSE.
 *********************************************************/
/*  Include file depends on local installation */
#include <SWI-Prolog.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pwd.h> 

/* general password database iterator returns all matches
   for any specified pattern.  This is more expensive than simply
   matching UID and LOGIN, so a separate function getpw is used to
   match them and return unique and simple results.  */

static struct passwd * passwd_copy(struct passwd *p) { 
  struct passwd *c = (struct passwd *) malloc (sizeof(struct passwd)); 
  if (p->pw_name) { 
    c->pw_name = (char *) malloc(strlen(p->pw_name)+1); 
    strcpy(c->pw_name,p->pw_name); 
  } else { 
    c->pw_name=NULL; 
  } 
  if (p->pw_passwd) { 
    c->pw_passwd = (char *) malloc(strlen(p->pw_passwd)+1); 
    strcpy(c->pw_passwd,p->pw_passwd); 
  } else { 
    c->pw_passwd=NULL; 
  } 
  c->pw_uid = p->pw_uid; 
  c->pw_gid = p->pw_gid; 
  if (p->pw_age) { 
    c->pw_age = (char *) malloc(strlen(p->pw_age)+1); 
    strcpy(c->pw_age,p->pw_age); 
  } else { 
    c->pw_age=NULL; 
  } 
  if (p->pw_comment) { 
    c->pw_comment = (char *) malloc(strlen(p->pw_comment)+1); 
    strcpy(c->pw_comment,p->pw_comment); 
  } else { 
    c->pw_comment=NULL; 
  } 
  if (p->pw_gecos) { 
    c->pw_gecos = (char *) malloc(strlen(p->pw_gecos)+1); 
    strcpy(c->pw_gecos,p->pw_gecos); 
  } else { 
    c->pw_gecos=NULL; 
  } 
  if (p->pw_dir) { 
    c->pw_dir = (char *) malloc(strlen(p->pw_dir)+1); 
    strcpy(c->pw_dir,p->pw_dir); 
  } else { 
    c->pw_dir=NULL; 
  } 
  if (p->pw_shell) { 
    c->pw_shell = (char *) malloc(strlen(p->pw_shell)+1); 
    strcpy(c->pw_shell,p->pw_shell); 
  } else { 
    c->pw_shell=NULL; 
  } 
  return c; 
} 

static void passwd_free (struct passwd *p) { 
   if (p->pw_name) free(p->pw_name); 
   if (p->pw_passwd) free(p->pw_passwd); 
   if (p->pw_age) free(p->pw_age); 
   if (p->pw_comment) free(p->pw_comment); 
   if (p->pw_gecos) free(p->pw_gecos); 
   if (p->pw_dir) free(p->pw_dir); 
   if (p->pw_shell) free(p->pw_shell); 
   free(p); 
} 

struct pw_listelem { 
  struct pw_listelem *next; 
  struct passwd *pw; 
} ;

struct pw_context { 
  struct pw_listelem *head, *tail; 
} ; 

static void pw_context_free(struct pw_context *c) { 
  struct pw_listelem *p; 
  while (c->head) { 
    p = c->head; 
    c->head = c->head->next; 
    passwd_free(p->pw); free(p); 
  } 
} 

foreign_t
pl_passwd(term_t name, 
      term_t passwd, 
      term_t uid, 
      term_t gid, 
      term_t gecos,
      term_t dir, 
      term_t shell, 
      foreign_t handle)
{
  struct pw_context *ctxt; 
  struct passwd *pw; 
  struct pw_listelem *p; 
  char *name_chars = NULL; 
  char *passwd_chars = NULL; 
  int uid_val = -1;
  int gid_val = -1; 
  char *gecos_chars = NULL; 
  char *dir_chars = NULL; 
  char *shell_chars = NULL; 
  if (PL_is_atom(name)) { 
    if (PL_get_atom_chars(name,&name_chars)) { 
      switch( PL_foreign_control(handle) )
      { case PL_FIRST_CALL:
	/* while this SEEMS like a good idea, it has the side 
           effect of only looking at the first match record; 
           this deletes the second value of the match record from 
           a secondary name service. This must be fixed! */ 
	  pw = getpwnam(name_chars);
	  if (pw
	   && PL_unify_atom_chars(passwd, pw->pw_passwd)
	   && PL_unify_integer(uid,pw->pw_uid)
	   && PL_unify_integer(gid,pw->pw_gid)
	   && PL_unify_atom_chars(gecos, pw->pw_gecos)
	   && PL_unify_atom_chars(dir, pw->pw_dir)
	   && PL_unify_atom_chars(shell, pw->pw_shell)) 
	    PL_succeed;
	  else 
	    PL_fail; 
	  break; 
        case PL_REDO:
        case PL_CUTTED: 
        default: 
	  PL_fail; 
      } 
    } else { 
      PL_fail; 
    } 
  } else if (PL_is_integer(uid)) { 
    if (PL_get_integer(uid, (int *)&uid_val)) {
      switch( PL_foreign_control(handle) )
      { case PL_FIRST_CALL:
	/* while this SEEMS like a good idea, it has the side 
           effect of only looking at the first match record; 
           this deletes the second value of the match record from 
           a secondary name service. This must be fixed! */ 
          pw = getpwuid(uid_val);
	  if (pw
	   && PL_unify_atom_chars(name, pw->pw_name)
	   && PL_unify_atom_chars(passwd, pw->pw_passwd)
	   && PL_unify_integer(gid,pw->pw_gid)
	   && PL_unify_atom_chars(gecos, pw->pw_gecos)
	   && PL_unify_atom_chars(dir, pw->pw_dir)
	   && PL_unify_atom_chars(shell, pw->pw_shell)) 
	    PL_succeed;
	  else 
	    PL_fail; 
	  break; 
        case PL_REDO:
        case PL_CUTTED: 
        default: 
	  PL_fail; 
      } 
    } else { 
      PL_fail; 
    } 
  } else { 
    switch( PL_foreign_control(handle) )
    { case PL_FIRST_CALL:

	/* first, avoid iteration if all is already lost */ 
	if (!PL_is_variable(name) 
	 && !PL_get_atom_chars(name,&name_chars))
	  PL_fail; 
	if (!PL_is_variable(passwd) 
	 && !PL_get_atom_chars(passwd,&passwd_chars)) 
	  PL_fail; 
	if (!PL_is_variable(uid) 
	 && !PL_get_integer(uid,&uid_val)) 
	  PL_fail; 
	if (!PL_is_variable(gid) 
	 && !PL_get_integer(gid,&gid_val))
	  PL_fail; 
	if (!PL_is_variable(gecos) 
	 && !PL_get_atom_chars(gecos,&gecos_chars)) 
	  PL_fail; 
	if (!PL_is_variable(dir) 
	 && !PL_get_atom_chars(dir,&dir_chars)) 
	  PL_fail; 
	if (!PL_is_variable(shell) 
	 && !PL_get_atom_chars(shell,&shell_chars)) 
	  PL_fail; 

	/* make a list of all matching entries, as of now */ 
	ctxt = (struct pw_context *) malloc (sizeof(struct pw_context)); 
	ctxt->head=ctxt->tail=NULL; 
	setpwent(); 
	while ((pw=getpwent())) { 
	  if ((!name_chars || strcmp(name_chars,pw->pw_name)==0) 
	   && (!passwd_chars || strcmp(passwd_chars,pw->pw_passwd)==0) 
	   && (uid_val==-1 || uid_val==pw->pw_uid) 
	   && (gid_val==-1 || gid_val==pw->pw_gid) 
	   && (!gecos_chars || strcmp(gecos_chars,pw->pw_gecos)==0) 
	   && (!dir_chars || strcmp(dir_chars,pw->pw_dir)==0) 
	   && (!shell_chars || strcmp(shell_chars,pw->pw_shell)==0)) {
	    p = (struct pw_listelem *) malloc(sizeof(struct pw_listelem)); 
	    p->pw = passwd_copy(pw); 
	    if (!ctxt->head) ctxt->head=ctxt->tail=p; 
	    else { ctxt->tail->next = p;  ctxt->tail=p; } 
	    p->next = NULL; 
	  } 
	} 

	/* now we have a linked list of matches; play them back
	   and simultaneously blow them away! */ 

	if (ctxt->head) { 
	  pw = ctxt->head->pw;
	  if( PL_unify_atom_chars(name, pw->pw_name)
	   && PL_unify_atom_chars(passwd, pw->pw_passwd)
	   && PL_unify_integer(uid,pw->pw_uid)
	   && PL_unify_integer(gid,pw->pw_gid)
	   && PL_unify_atom_chars(gecos, pw->pw_gecos)
	   && PL_unify_atom_chars(dir, pw->pw_dir)
	   && PL_unify_atom_chars(shell, pw->pw_shell)) {
	    p = ctxt->head; 
	    ctxt->head = ctxt->head->next; 
	    passwd_free(p->pw); free(p); 
	    PL_retry_address(ctxt);
	  } else { 
	    pw_context_free(ctxt); 
	    free(ctxt); 
	    PL_fail; 
	  } 
	} else { 
	  pw_context_free(ctxt); 
	  free(ctxt); 
	  PL_fail; 
	} 

      case PL_REDO:

	/* next element is always the first element of the list */ 

	ctxt = PL_foreign_context_address(handle); 
	if (ctxt->head) { 
	  pw = ctxt->head->pw;
	  if( PL_unify_atom_chars(name, pw->pw_name)
	   && PL_unify_atom_chars(passwd, pw->pw_passwd)
	   && PL_unify_integer(uid,pw->pw_uid)
	   && PL_unify_integer(gid,pw->pw_gid)
	   && PL_unify_atom_chars(gecos, pw->pw_gecos)
	   && PL_unify_atom_chars(dir, pw->pw_dir)
	   && PL_unify_atom_chars(shell, pw->pw_shell)) {
	    p = ctxt->head; 
	    ctxt->head = ctxt->head->next; 
	    passwd_free(p->pw); free(p); 
	    PL_retry_address(ctxt);
	  } else { 
	    pw_context_free(ctxt); 
	    free(ctxt); 
	    PL_fail; 
	  } 
	} else { 
	  pw_context_free(ctxt); 
	  free(ctxt); 
	  PL_fail; 
	} 

      case PL_CUTTED:

	/* cut context */ 
	ctxt = PL_foreign_context_address(handle); 
	pw_context_free(ctxt); 
	free(ctxt); 
	PL_succeed;
    }
  } 
  PL_fail; 
} 

install_t
pl_passwd_install()
{ 
  PL_register_foreign("passwd", 7, pl_passwd, PL_FA_NONDETERMINISTIC);
}
